from jto.dataclass_generator import ClassesTemplate, FieldTemplate
from jto.undefined_field import Undefined

from src.jto.dataclass_generator import ClassTemplate


def test_parse_empty_dict():
    data = {}

    test_classes = ClassesTemplate()
    test_classes.build_classes('Response', data)
    expected_classes = ClassesTemplate(classes=[ClassTemplate(class_name='Response', class_fields=[])])
    assert str(test_classes) == str(expected_classes)

    classes_str = test_classes.build_classes_string()
    assert classes_str == "@dataclass\nclass Response:\n    "


def test_empty_list_value():
    data = {'var': []}

    test_classes = ClassesTemplate()
    test_classes.build_classes('Response', data)
    expected_classes = ClassesTemplate(classes=[
        ClassTemplate(class_name='Response', class_fields=[
            FieldTemplate(field_name='var', field_type="List[<class 'object'>]",
                          json_field_name='var', default_value=Undefined.__name__, required=False)
        ])
    ])
    assert str(test_classes) == str(expected_classes)

    classes_str = test_classes.build_classes_string()
    assert classes_str == "@dataclass\nclass Response:\n    var: List[<class 'object'>] = " \
                          "field(default=Undefined, metadata={'name': 'var', 'required': False})"



from typing import TypeVar

from jto.json_builder import JsonBuilder
from jto.json_parser import JsonParser


class JTOConverter:
    T = TypeVar('T')

    @classmethod
    def from_json(cls, dataclass_type: T, json_data: dict) -> T:
        result = JsonParser.parse_json(dataclass_type, json_data)
        return result

    @classmethod
    def to_json(cls, dataclass_obj, drop_nones: bool = False) -> dict:
        result = JsonBuilder.build_json(dataclass_obj, drop_nones)
        return result



from jto.dataclass_generator import ClassesTemplate

data = {
  "Id": 12345,
  "Customer": "John Smith",
  "Quantity": 1,
  "Price": 10.00
}

test_classes = ClassesTemplate()
test_classes.build_classes('Request', data)
print(test_classes.build_classes_string())





# from dataclasses import dataclass, field, asdict
#
# class Undefined:
#     pass
#
# @dataclass
# class Test:
#     f1: str = field(default=Undefined, metadata={'name': 'f1', 'required': False})
#
# result = asdict(Test())
# print(result, result['f1'] == Undefined)



from dataclasses import is_dataclass, asdict

from jto.undefined_field import Undefined


class JsonBuilder:
    @classmethod
    def build_json(cls, dataclass_obj, drop_nones: bool) -> dict:
        if not is_dataclass(dataclass_obj):
            raise ValueError(f'Dataclass type object expected, but received "{str(type(dataclass_obj))}"')

        result = asdict(dataclass_obj)
        result = cls.__drop_undefined(result)
        if drop_nones:
            result = cls.__drop_nones(result)
        return result

    @classmethod
    def __drop_undefined(cls, original_dict: dict) -> dict:
        result_dict = {}
        for key, value in original_dict.items():
            if value != Undefined:
                if isinstance(value, dict):
                    result_dict[key] = cls.__drop_undefined(value)
                elif isinstance(value, (list, set, tuple)):
                    result_dict[key] = cls.__drop_undefined_in_list(value)
                else:
                    result_dict[key] = value
        return result_dict

    @classmethod
    def __drop_undefined_in_list(cls, original_list: list) -> list:
        result_list = []
        for value in original_list:
            if value != Undefined:
                if isinstance(value, dict):
                    result_list.append(cls.__drop_undefined(value))
                elif isinstance(value, (list, set, tuple)):
                    result_list.append(cls.__drop_undefined_in_list(value))
                else:
                    result_list.append(value)
        return result_list

    @classmethod
    def __drop_nones(cls, original_dict: dict) -> dict:
        result_dict = {}
        for key, value in original_dict.items():
            if isinstance(value, dict):
                result_dict[key] = cls.__drop_nones(value)
            elif isinstance(value, (list, set, tuple)):
                result_dict[key] = cls.__drop_nones_in_list(value)
            elif value is not None:
                result_dict[key] = value
        return result_dict

    @classmethod
    def __drop_nones_in_list(cls, original_list: list) -> list:
        result_list = []
        for value in original_list:
            if isinstance(value, dict):
                result_list.append(cls.__drop_nones(value))
            elif isinstance(value, (list, set, tuple)):
                result_list.append(cls.__drop_nones_in_list(value))
            else:
                result_list.append(value)
        return result_list



from dataclasses import dataclass, field
from typing import List

import pytest
from jto import JTOConverter


def test_convert_empty_dict():
    data = {}

    @dataclass
    class Test:
        f1: str = field(default=None, metadata={'name': 'f1', 'required': False})

    dataclass_object = JTOConverter.from_json(Test, data)
    assert dataclass_object == Test()

    json_object = JTOConverter.to_json(dataclass_object, drop_nones=True)
    assert json_object == {}


def test_convert_empty_list():
    data = {"f1": []}

    @dataclass
    class Test:
        f1: List[str] = field(default=None, metadata={'name': 'f1', 'required': False})

    dataclass_object = JTOConverter.from_json(Test, data)
    assert dataclass_object == Test(f1=[])


def test_convert_dict_missing_required_field():
    data = {}

    @dataclass
    class Test:
        f1: str = field(default=None, metadata={'name': 'f1', 'required': True})

    with pytest.raises(ValueError, match='Required field "f1" not found in the data "{}"'):
        JTOConverter.from_json(Test, data)


def test_convert_dict_field_with_null_value():
    data = {"f1": None}

    @dataclass
    class Test:
        f1: str = field(default=None, metadata={'name': 'f1', 'required': False})

    dataclass_object = JTOConverter.from_json(Test, data)
    assert dataclass_object == Test()


def test_convert_dict_with_unexpected_values():
    data = {"one": 1}

    @dataclass
    class Test:
        f1: str = field(default=None, metadata={'name': 'f1', 'required': False})

    dataclass_object = JTOConverter.from_json(Test, data)
    assert dataclass_object == Test()


def test_convert_value_with_unexpected_type():
    data = {"f1": 1}

    @dataclass
    class Test:
        f1: str = field(default=None, metadata={'name': 'f1', 'required': False})

    with pytest.raises(TypeError, match='Expected value type is "<class \'str\'>", but received "<class \'int\'>"'):
        JTOConverter.from_json(Test, data)


def test_convert_dataclass_to_json():
    @dataclass
    class Request:
        Id: int = field(default=None, metadata={'name': 'Id', 'required': False})
        Customer: str = field(default=None, metadata={'name': 'Customer', 'required': False})
        Quantity: int = field(default=None, metadata={'name': 'Quantity', 'required': False})
        Price: float = field(default=None, metadata={'name': 'Price', 'required': False})

    dataclass_object = Request(1, 'aaa', 2, 3.33)

    json_object = JTOConverter.to_json(dataclass_object, drop_nones=True)

    expected_json = {'Id': 1, 'Customer': 'aaa', 'Quantity': 2, 'Price': 3.33}
    assert json_object == expected_json




from dataclasses import is_dataclass, fields, Field
from typing import get_origin, get_args, TypeVar


class JsonParser:
    T = TypeVar('T')

    @classmethod
    def parse_json(cls, dataclass_type: T, json_data: dict) -> T:
        if not is_dataclass(dataclass_type):
            raise ValueError(f'Dataclass type expected, but received "{str(type(dataclass_type))}"')
        result = cls.__parse_dict(dataclass_type, json_data)
        return result

    @classmethod
    def __parse_dict(cls, dataclass_type: T, json_data: dict):
        dataclass_obj: dataclass_type = dataclass_type()
        for dataclass_field in fields(dataclass_type):
            cls.__parse_dict_item(dataclass_field, json_data, dataclass_obj)
        return dataclass_obj

    @classmethod
    def __parse_dict_item(cls, class_field: Field, json_data: dict, result_class):
        for key, value in json_data.items():
            if class_field.metadata['name'] == key:
                if value is None:
                    setattr(result_class, key, None)
                else:
                    if is_dataclass(class_field.type):
                        setattr(result_class, key, cls.__parse_dict(class_field.type, value))

                    elif str(class_field.type)[:11] == 'typing.List':
                        setattr(result_class, key, cls.__parse_list(class_field, value))

                    else:
                        if class_field.type != type(value):
                            raise TypeError(f'Expected value type is "{str(class_field.type)}", '
                                            f'but received "{str(type(value))}"')
                        setattr(result_class, key, value)
                return

        if class_field.metadata['required']:
            raise ValueError(f'Required field "{class_field.name}" not found in the data "{json_data}"')

    @classmethod
    def __parse_list(cls, class_field, json_value: list):
        if get_origin(class_field.type) != list or get_args(class_field.type) == ():
            raise ValueError(f'class_field type "{str(class_field.type)}" is not a supported list. '
                             f'Change type to List[YourClass]')
        if type(json_value) != list:
            raise ValueError(f'json_value type "{str(type(json_value))}" is not a list.')

        class_type = get_args(class_field.type)[0]

        items = []
        if is_dataclass(class_type):
            for item in json_value:
                final_item = cls.__parse_dict(class_type, item)
                items.append(final_item)
        else:
            items.extend(json_value)

        return items



from dataclasses import dataclass, field

from jto.undefined_field import Undefined


def to_camel_case(text):
    s = text.replace("-", " ").replace("_", " ")
    s = s.split()
    if len(text) == 0:
        return text
    return ''.join(i.capitalize() for i in s)


@dataclass
class FieldTemplate:
    field_name: str
    field_type: str
    json_field_name: str
    default_value: any = Undefined.__name__
    required: bool = False

    def build_field_string(self) -> str:
        field_string = f"{self.field_name}: {self.field_type} = field(default={str(self.default_value)}, " \
                       f"metadata={{'name': '{self.json_field_name}', 'required': {str(self.required)}}})"
        return field_string


@dataclass
class ClassTemplate:
    class_name: str
    class_fields: list = field(default_factory=list)

    def build_class_string(self):
        class_string = f'@dataclass\nclass {self.class_name}:\n'
        class_field_strings = [cls_field.build_field_string() for cls_field in self.class_fields]
        fields_string = '\n    '.join(class_field_strings)
        return class_string+'    '+fields_string


@dataclass
class ClassesTemplate:
    classes: list = field(default_factory=list)

    def build_classes_string(self):
        class_strings = [cls_temp.build_class_string() for cls_temp in self.classes]
        result_string = '\n\n'.join(class_strings)
        return result_string

    def _parse_dict(self, result_class, key, value, is_list_item: bool = False):
        cls_name = to_camel_case(key)
        self.build_classes(cls_name, value)
        if is_list_item:
            result_class.class_fields.append(FieldTemplate(key, f'List[{cls_name}]', key))
        else:
            result_class.class_fields.append(FieldTemplate(key, cls_name, key))

    def _parse_list(self, result_class, key, value):
        if len(value) == 0:
            result_class.class_fields.append(FieldTemplate(key, f'List[{object}]', key))
        else:
            list_element = value[0]
            if type(list_element) == dict:
                self._parse_dict(result_class, key, list_element, True)
            else:
                result_class.class_fields.append(FieldTemplate(key, f'List[{type(value).__qualname__}]', key))

    def build_classes(self, class_name: str, json_data: dict):
        result_class = ClassTemplate(class_name)
        for key, value in json_data.items():
            if type(value) == dict:
                self._parse_dict(result_class, key, value, False)
            elif type(value) == list:
                self._parse_list(result_class, key, value)
            else:
                result_class.class_fields.append(FieldTemplate(key, type(value).__qualname__, key))
        self.classes.append(result_class)
